--------------------------------------------------------------------------------
-- From Deriving to Multirec
--------------------------------------------------------------------------------

{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE ConstraintKinds        #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE MultiParamTypeClasses  #-}
{-# LANGUAGE OverlappingInstances   #-}
{-# LANGUAGE UndecidableInstances   #-}

module Conversion.D2M where


-- Type conversion
import qualified Library.Deriving as D
import qualified Library.Multirec as M
import Auxiliary.Utils
import GHC.Exts (Constraint)

data Void where

type family D2MFam (ts :: [i]) :: M.Un * i
type instance D2MFam '[]       = M.K Void
type instance D2MFam (t ': ts) = M.SUM (M.TAG (D2M (D.Rep t)) t) (D2MFam ts)

type family D2M (u :: D.Un *) :: M.Un * i
type instance D2M D.UU              = M.U
type instance D2M ((D.:+:)  a b)    = M.SUM  (D2M a) (D2M b)
type instance D2M ((D.:**:) a b)    = M.PROD (D2M a) (D2M b)
type instance D2M (D.MM c a)        = D2M a
type instance D2M (D.KK (D.R so) r) = M.I r
type instance D2M (D.KK D.P a)      = M.K a
type instance D2M (D.KK D.U a)      = M.K a


-- Value conversion
class Convert (u :: D.Un *) where
  d2m :: D.In u p -> M.In (D2M u) M.I0 t
  m2d :: M.In (D2M u) M.I0 t -> D.In u p

instance Convert D.UU where
  d2m _ = M.In_U
  m2d _ = D.U1

instance (Convert a) => Convert (D.MM c a) where
  d2m (D.M1 a) = d2m a
  m2d a        = D.M1 (m2d a)

instance Convert (D.KK D.P a) where
  d2m (D.K1 a)   = M.In_K a
  m2d (M.In_K a) = D.K1 a

instance Convert (D.KK (D.R so) r) where
  d2m (D.K1 a)          = M.In_I (M.I0 a)
  m2d (M.In_I (M.I0 a)) = D.K1 a

instance Convert (D.KK D.U a) where
  d2m (D.K1 a)   = M.In_K a
  m2d (M.In_K a) = D.K1 a

instance (Convert a, Convert b) => Convert ((D.:+:) a b) where
  d2m (D.L1 a)   = M.In_L (d2m a)
  d2m (D.R1 a)   = M.In_R (d2m a)
  m2d (M.In_L a) = D.L1 (m2d a)
  m2d (M.In_R a) = D.R1 (m2d a)

instance (Convert a, Convert b) => Convert ((D.:**:) a b) where
  d2m ((D.:*:) a b) = M.In_P  (d2m a) (d2m b)
  m2d (M.In_P  a b) = (D.:*:) (m2d a) (m2d b)


-- Embedding conversion
-- Multirec phi-like families from type-level lists of datatypes
data (:@:) (ts :: [*]) (t :: *) where
  This ::             (t  ': ts) :@: t
  That :: ts :@: t -> (t' ': ts) :@: t

instance M.El ((:@:) (t1 ': ts)) t1 where
  proof = This

instance (M.El ((:@:) ts) t) => M.El ((:@:) (t1 ': ts)) t where
  proof = That M.proof

instance (FamConstrs ts) => M.Fam ((:@:) ts) where
  type PF ((:@:) ts) = D2MFam ts

  from This     x = M.In_L (M.In_T (d2m (D.from x)))
  from (That k) x = M.In_R (M.from k x)

  to This     (M.In_L (M.In_T x)) = D.to (m2d x)
  to (That k) (M.In_R x)          = M.to k x

-- Compute the required constraints for conversion
type family FamConstrs (ts :: [k]) :: Constraint
type instance FamConstrs '[]       = ()
type instance FamConstrs (t ': ts) = ( D.Generic t
                                     , Convert (D.Rep t)
                                     , M.Fam ((:@:) ts)
                                     , FamConstrs ts)
